#                                                         -*- Perl -*-
# jargon-fpw - FreePWING script for The Jargon File
#
# $Id: jargon-fpw.pl 1.2 2001/06/11 18:46:11 satomii Exp $
# Copyright (C) 2000-2001 <satomi@ring.gr.jp>
#
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
# This script is not so neat as you see - that's because I am not a skilled
# Perl programmer. Please re-write if you don't like it :-)
#

use FreePWING::FPWUtils::FPWParser;
use Text::Tabs;
use strict;

use subs qw(
	&compare_insensitive($$)
	&fpwerr($)
	&preprocess()
	&write_words()
	&write_copyright()
	&write_menu()
	&write_cross_reference($$$$)
	&write_preformat($$)
	&write_body($@)
	&get_block($@)
);

my $VERSION = '1.2';
my $COPYRIGHT = '(C) 2000-2001 satomi';

my($fpwtext, $fpwheading, $fpwword2, $fpwmenu, $fpwcopyright);
my(%words, %menues);


##-----------------------------------------------------------------------
##  MAIN ROUTINE

initialize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'menu' => \$fpwmenu,
		'copyright' => \$fpwcopyright);

my $menu_temp_file = "$work_directory/menu.tmp";
my $entry_temp_file = "$work_directory/jargon.tmp";

open(MENU, "+>$menu_temp_file") ||
		die "Failed to create a temp file \'$menu_temp_file\'.";
open(ENTRY, "+>$entry_temp_file") ||
		die "Failed to create a temp file \'$entry_temp_file\'.";

&preprocess();

seek(MENU, 0, 0);
seek(ENTRY, 0, 0);

&write_words();
&write_copyright();
&write_menu();

finalize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'menu' => \$fpwmenu,
		'copyright' => \$fpwcopyright);

close(MENU);
close(ENTRY);
unlink($menu_temp_file);
unlink($entry_temp_file);

exit;


##-----------------------------------------------------------------------
##  HELPER FUNCTIONS

#
# int compare_insensitive(char *a, char *b)
# compare routine for jargon word entries.
#
sub compare_insensitive($$)
{
	my($cmp1, $cmp2);

	if (defined($a) && defined($b)) {
		($cmp1, $cmp2) = ($a, $b);
	} else {
		($cmp1, $cmp2) = @_;
	}

	$cmp1 =~ tr/a-z/A-Z/;
	$cmp2 =~ tr/a-z/A-Z/;
	$cmp1 =~ s/[^A-Z0-9@]//g;
	$cmp2 =~ s/[^A-Z0-9@]//g;
	$cmp1 cmp $cmp2;
}

#
# char *fpwerr(void *obj)
# returns the FreePWING error message surrounded by ().
#
sub fpwerr($)
{
	my $obj = shift(@_);
	my $err = $obj->error_message();
	if ($err) {
		return(" ($err)");
	}
	return("");
}

##-----------------------------------------------------------------------
##  PREPROCESSOR

#
# void preprocess(void)
# read input file(s) and save its info into the corresponding
# temporary file.
#
sub preprocess()
{
	my($menu_count, $entry_count, $entry_type) = (0, 0, -1);
	my($line, $word, $heading_level);

	print "pre-processing the input file...\n";

	while (<>) {
		chomp();
		$_ = Text::Tabs::expand($_);
		s/([^\s]*)\s+$/$1/;

		if (s/^#=+([^=]+)=+#$/\1/) {
			#
			# BOF or EOF marker.
			#
			s/^ +(.*) +$/\1/;

			if (/ENDS HERE$/) {
				# jarg422 will not reach here.
				$entry_type = -1;
				next;
			}

			if (0 < $menu_count) {
				print MENU "\n";
			}
			print MENU "0:$_";
			$entry_type = 0;
			$menu_count++;
			$menues{$_} = $menu_count;

		} elsif (/\s*:\(Lexicon Entries End Here\):/i) {
			#
			# fix for jarg422 or later.
			#
			$entry_type = -1;

		} elsif (s/^(:.*[^:]*:)\s*([\*\=\-]*)$/$1/
				|| s/^(The Jargon Lexicon)\s*(\**)$/$1/
				|| s/^(= [A-Z0] =)\s*(\=*)$/$1/
				#
				# jarg430 incorrectly formats the following entry possibly
				# because of a colon in the title.
				#
				#|| s/^:(TV Typewriters:.+)$/$1/
				|| s/^:(TV Typewriters:):*(\s*A Tale of Hackish Ingenuity)\s*(.*)$/$1$2/
				) {
			#
			# section heading.
			#
			if ($3) {
				# 'TV Typewriters' section, should be followed by '='.
				$line = $3;
				$heading_level = 2;

			} else {
				if ($2) {
					# [\*\=\-] is included in the same line.
					$line = $2;
				} else {
					# next line should be /^[\*\=\-]+$/.
					$line = <>;
					chomp($line);
				}

				# remove the sectiono marker.
				s/^:(.*[^:]*):$/$1/;

				if ($line =~ /^\*+$/) {
					$heading_level = 1;
				} elsif ($line =~ /^=+$/) {
					$heading_level = 2;
				} elsif ($line =~ /^\-+$/) {
					$heading_level = 3;
				} else {
					print MENU "\t:$_:\t$line";
					next;
				}
			}

			#
			# sorry, i don't like this way when viewing with Windows font
			# (but others may prefer).
			#
			s/`([^\']+)\'/\'$1\'/g;

			print MENU "\n$heading_level:$_";
			$entry_type = 0;
			$menu_count++;

			if (/^(TV Typewriters)/) {
				$menues{$1} = $menu_count;
				if ($line) {
					print MENU "\t$line";
				}
			} else {
				$menues{$_} = $menu_count;
			}

		} elsif (/^(?:   )?:([^:]+):+ *(.*)/) {		# word entry starts.
			$word = $1;
			next if ($word eq '(Lexicon Entries End Here)');
			$word =~ s/^ +//;
			$word =~ s/ +$//;
			print ENTRY "\n$word\t$2";
			$entry_type = 1;

			# store the word index into an associative array.
			# this array will later be used to make cross-references.
			$entry_count++;
			$words{$word} = $entry_count;

		} elsif ($entry_type == 1) {
			#
			# word entry continues.
			#
			print ENTRY "\t$_";

		} elsif ($entry_type == 0) {
			#
			# menu entry continues.
			#
			if (s/\s*\#\=+ THE JARGON FILE ENDS HERE$//) {
				# fix for jarg422.
				$entry_type = -1;
				print MENU "\t$_\n";
				last;
			}
			print MENU "\t$_";
		}
	}
}


##-----------------------------------------------------------------------
##  COMMON WRITING FUNCTIONS

#
# void write_cross_reference(obj, char *key, char *text, bool bracket)
# write a word enclosed in cross-reference tags. ignores any error
# when adding the tags (but dies if the word could not be written).
#
sub write_cross_reference($$$$)
{
	my($obj, $key, $text, $bracket) = @_;
	my($tag) = '';

	$key =~ s/^(TV Typewriters).*$/\1/;

	if (defined($words{$key})) {
		$tag = sprintf("JARG%04d", $words{$key});
	} elsif (defined($menues{$key})) {
		$tag = sprintf("MENU%04d", $menues{$key});
	}

	if ($bracket) {
		$obj->add_text("{")
			|| die "Cannot write bracket \'{\'" . &fpwerr($obj) . ".";
	}
	if ($tag ne '') {
		$obj->add_reference_start();
	}

	$obj->add_text($text)
		|| die "Cannot write text: $text" . &fpwerr($obj);

	if ($tag ne '') {
		$obj->add_reference_end($tag);
	}
	if ($bracket) {
		$obj->add_text("}")
			|| die "Cannot write bracket \'}\'" . &fpwerr($obj) . ".";
	}
}

#
# void write_preformat(obj, char *line)
# write a pre-formatted part, where some symbol characters are
# represented using gaiji.
#
sub write_preformat($$)
{
	my($obj) = shift(@_);
	my(@chars) = unpack('C*', $_[0]);
	my($str) = '';
	my($char);

	while (0 < scalar(@chars)) {
		$char = shift(@chars);
		if (0xd7 == $char || 0x5c == $char) {
			if ('' ne $str) {
				$obj->add_text($str)
					|| die "Cannot write body text: $str" . &fpwerr($obj);
				$str = '';
			}
			$obj->add_half_user_character(sprintf("%02x", $char))
				|| die "Cannot add gaiji for \xd7" . &fpwerr($obj) . ".";
		} else {
			$str .= pack('C', $char);
		}
	}

	if ('' ne $str) {
		$obj->add_text($str)
			|| die "Cannot write body text: $str" . &fpwerr($obj);
	}
}

#
# void write_body(obj, char *args[])
# write the description part. text-formatting routine is not perfect,
# so simplify this code if you don't like the result.
#
sub write_body($@)
{
	my($obj, @args) = @_;
	my($i, @lines, @parts, $line, $part) = (0);

	while ($i < @args) {
		# retrieve one block, which should have the same
		# indent level.
		my($indent_level, $count, $line, $pre_format) = &get_block($i, @args);
		if (!$count) {
			$i++;
			next;
		} else {
			$i += $count;
		}

		# set indent.
		$obj->add_indent_level($indent_level)
			|| die "Failed to set indent" . &fpwerr($obj) . ".";

		# divide lines and store them into an array.
		$_ = $line;
		s/([^a-z ]) ([1-9]\. )/\1\n\2/g;
		s/`([^']*)'/'\1'/g;
		@lines = split(/\n/);

		foreach $line (@lines) {
			next if ($line =~ /^$/);

			# divide cross-references (if any).
			$line =~ s/({[^{}]+})/\n\1\n/g;
			@parts = split(/\n/, $line);

			$obj->add_nowrap_start() if ($pre_format);

			foreach $part (@parts) {
				next if ($part =~ /^$/);
				if ($part =~ s/^{([^{}]+)}$/\1/) {
					&write_cross_reference($obj, $part, $part, 1);
				} elsif ($pre_format) {
					&write_preformat($obj, $part);
				} else {
					$obj->add_text($part)
						|| die "Cannot write body text: $part" . &fpwerr($obj);
				}
			}

			# add a new line at the end of each line.
			$obj->add_nowrap_end() if ($pre_format);
			$obj->add_newline()
				|| die "Failed to add newline" . &fpwerr($obj) . ".";
		}
	}
}

#
# void *get_block(int index, char *args[])
# retrieves one 'block' from the arguments (lines) at the specified index.
# return values are in array context and contains the indent level, line
# count, block text and a flag whether this block is preformatted or not.
#
sub get_block($@)
{
	my($start) = shift(@_);
	my(@args) = @_;
	my($indent_level, $space_count, $pre_format) = (2, 0, 0);
	my($i, @lines, $line, $line_count);

	# checks for the format of the line very roughly. some codes might
	# need to be added to tidy up more neatly.
	$_ = $args[$start];

	if (/^           /) {
		$space_count = 5;
		$pre_format = 1;
	} elsif (/^(          )/ || /^(         )/) {
		$space_count = length($1);
	} elsif (/^          >/) {
		$space_count = 10;
		$pre_format = 1;
	} elsif (/^( +)[A-Z][:\.] /) {
		$space_count = length($1);
		if ($space_count < 4) {
			$indent_level = 3;
		}
	} elsif (/^( +)\* /) {
		$space_count = length($1);
		if ($space_count < 4) {
			$indent_level = 3;
		}
	} elsif (/^(     )/ || /^(    )/) {
		$space_count = length($1);
	} elsif (/^( +)[1-9]\. /
			|| /^( +)[1-9][0-9]*\. /) {
		$space_count = length($1);
		$indent_level = 3;
	} elsif (/^:[^:]+:$/) {
		$pre_format = 1;
	}

	# determine the indent level.
	if ($indent_level == 2) {
		if (10 <= $space_count) {
			$indent_level = 6;
		} elsif (8 <= $space_count) {
			$indent_level = 5;
		} elsif (6 <= $space_count) {
			$indent_level = 4;
		} elsif (4 <= $space_count) {
			$indent_level = 3;
		}
	}

	# sometimes the next line wouldn't be indented.
	if ($start + 1 < @args && $args[$start + 1] =~ /^[^ ]/) {
		$space_count = 0;
	}

	# check to see if the format should be preserved.
	if ($pre_format == 0 && 2 < $indent_level) {
		if (/^ +\* (.*)/
				|| /^ +[A-Z][:\.] /
				|| /^ +[1-9][0-9]*\. /) {
			if (length($_) < 40) {
				$pre_format = 1;
			}
		} elsif (length($_) < 50) {
			if ($start + 1 < @args
					&& $args[$start + 1] =~ /^( +)/
					&& length($1) == $space_count + 6) {
				# such as pos list, face-mark list etc.
				push(@lines, $_);
				$start = @args;		# to skip the for loop.
			} else {
				$pre_format = 1;
			}
		} elsif (/^[^a-zA-Z]+$/ || /[^a-zA-Z0-9\s\"\#]{10}/) {
			$pre_format = 1;

		} else {
			my($chk) = $_;
			$chk =~ s/^ +//;
			if ($chk =~ /^[^a-zA-Z0-9\({\"`<\[]/) {
				$pre_format = 1;

			# for individual adjustments...
			} elsif ($chk =~ /^high low bits bits 0000/) {
				$pre_format = 1;	# :bit-paried keyboard:
			} elsif ($chk =~ /\"The disk heads just crashed\.\"/) {
				$pre_format = 1;
			} elsif ($chk =~ /monstrosity  /) {
				$pre_format = 1;
			} elsif ($chk =~ /In <[a-zA-Z0-9@\.]+>.+:$/) {
				$pre_format = 1;	# e-mail quotes.
			} elsif ($chk =~ /I've seen that term used on/) {
				$pre_format = 1;
			} elsif ($chk =~ /o\-+\)\|\|\(\-+/) {
				$pre_format = 1;	# ascii art.
			}
		}
	}

	for ($i = $start; $i < @args; $i++) {
		$_ = $args[$i];

		# blank lines should be preserved in the pre-formatted blocks,
		# or exit from this loop.
		if (/^ *$/) {
			$_ = "\n \n";
			if ($pre_format == 0) {
				push(@lines, $_);
				last;
			}
		# check for the indent level.
		} elsif (s/^( +)//) {
			if (0 < $space_count) {
				last if (0 < length($1) && length($1) < $space_count);
				$_ = (' ' x (length($1) - $space_count)) . $_;
				if (/^[A-Z][:\.] /) {
					$space_count += 3;
				} elsif (/^([1-9][0-9]*\. )/) {
					$space_count += length($1);
				} elsif (/^\* /) {
					$space_count += 2;
				}
			} else {
				last if (3 < length($1));
				last if ($pre_format == 1 && 3 == length($1));
				s/^ +//;
			}
		} elsif (0 < $space_count) {
			last;
		}

		push(@lines, $_);
	}

	$line_count = @lines;

	# for a pre-formatted block, combine each lines with '\n'.
	if ($pre_format == 1) {
		$line = join("\n", @lines);
	# for a normal block, remove extra spaces.
	} else {
		$line = "@lines";
		$line =~ s/  +/ /g;
	}

	# sorry, i don't like these ways (but others may prefer).
	$line =~ s/``([^']+)''/\"\1\"/g;
	$line =~ s/``([^']*'m [^']*)''/\"\1\"/g;
	$line =~ s/``([^']*'s [^']*)''/\"\1\"/g;
	$line =~ s/``([^']*n't[^']*)''/\"\1\"/g;
	$line =~ s/``([^']+)'([^'])/'\1'\2/g;
	$line =~ s/`([^']+)'/'\1'/g;
	$line =~ s/`([0-9]+'s)'/'\1'/g;

	return($indent_level, $line_count, $line, $pre_format);
}


##-----------------------------------------------------------------------
##  GLOSSARY ENTRIES

#
# void write_words(void)
# writes each word entries to the dictionary.
#
sub write_words()
{
	my($word, @desc);
	my($heading_position, $text_position, $key);

	while (<ENTRY>) {
		next if (/^$/);
		chomp();
		($word, @desc) = split(/\t/, $_);

		if (!defined($words{$word})) {
			die "Cannot find tag index for word \'$word\'.";
		}
		print "adding $word...\n";

		# switch to the next entry.
		$fpwtext->new_entry()
			|| die "Unable to add a new entry", &fpwerr($fpwtext) . ".";
		$fpwheading->new_entry()
			|| die "Unable to add a new entry", &fpwerr($fpwheading) . ".";

		# set a cross-reference tag.
		$key = sprintf("JARG%04d", $words{$word});
		$fpwtext->add_tag($key)
			|| die "Cannot add cross-reference tag \'$key\'"
					. &fpwerr($fpwtext) . ".";

		# add search keyword.
		$heading_position = $fpwheading->entry_position();
		$text_position = $fpwtext->entry_position();
		$fpwword2->add_entry($word, $heading_position,
				$heading_file_name, $text_position, $text_file_name)
			|| die "Cannot add search word \'$word\'"
					. &fpwerr($fpwword2) . ".";

		# write the keyword.
		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($word)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			die "Cannot write keyword \'$word\'" . &fpwerr($fpwtext) . ".";
		}

		# write the heading.
		$fpwheading->add_text($word)
			|| die "Cannot write heading \'$word\'"
					. &fpwerr($fpwheading) . ".";

		# write the description.
		&write_body($fpwtext, @desc);
		$fpwtext->add_newline()
			|| die "Failed to add newline" . &fpwerr($fpwtext) . ".";
	}
}


##-----------------------------------------------------------------------
##  COPYORIGHT ENTRY

#
# void write_copyright(void)
# writes the copyright information and generated date. all errors (if
# any) will be simply ignored in this function.
#
sub write_copyright()
{
	my $str;

	print "writing copyright information...\n";

	$fpwcopyright->add_keyword_start();
	$fpwcopyright->add_text("The Jargon File, FreePWING Version");
	$fpwcopyright->add_keyword_end();
	$fpwcopyright->add_newline();
	$fpwcopyright->add_newline();

	$fpwcopyright->add_text("The Jargon File by Eric S. Raymond");
	$fpwcopyright->add_newline();
	$fpwcopyright->add_text("FreePWING (C) 2000 Motoyuki Kasahara");
	$fpwcopyright->add_newline();
	$fpwcopyright->add_text("jargon-fpw $COPYRIGHT");
	$fpwcopyright->add_newline();
	$fpwcopyright->add_newline();

	if (defined($ENV{'USER'})) {
		$str = $ENV{'USER'};
	} elsif (defined($ENV{'USERNAME'})) {
		$str = $ENV{'USERNAME'};
	} else {
		$str = '(unknown user)';
	}
	$fpwcopyright->add_text("This book was generated by $str on ");

	$str = localtime(time);
	$fpwcopyright->add_text("$str using jargon-fpw $VERSION.");
	$fpwcopyright->add_newline();
}


##-----------------------------------------------------------------------
##  MENU ENTRIES

#
# void write_menu(void)
# creates a menu for the jargon file.
#
sub write_menu()
{
	my(@lines) = <MENU>;
	my($i, $title, @desc, @word_list);
	my($index_start, $index_char, $top_char) = (0, '');

	print "making table of contents for the menu...\n";
	for ($i = 0; $i < scalar(@lines); $i++) {
		&write_menu_content($lines[$i]);
	}
	$fpwmenu->add_newline()
		|| die "Failed to add newline" . &fpwerr($fpwmenu) . ".";

	# write each menu description.
	for ($i = 0; $i < @lines; $i++) {
		($title, @desc) = split(/\t/, $lines[$i]);
		$title =~ s/^[0-9]://;

		if ($title =~ /^= . =$/) {
			if (!$index_start) {
				$index_start = $i;
			}
			next;
		}

		&add_menu_page($title);
		&write_body($fpwmenu, @desc);

		if ($i + 1 == @lines) {
			# the last entry does not have a blank line at the end.
			$fpwmenu->add_newline()
				|| die "Failed to add newline" . &fpwerr($fpwmenu) . ".";
		}

		&write_menu_footer($i, @lines);
	}

	# write word index pages.
	@word_list = sort compare_insensitive keys(%words);

	for ($i = 0; $i < @word_list; $i++) {
		$word_list[$i] =~ /^[^a-zA-Z0-9@]*([a-zA-Z0-9@])/;
		$top_char = $1;

		if ($index_char eq '') {
			$index_char = '0';
			$title = '= 0 =';
		} elsif ($top_char =~ /[a-zA-Z]/
				&& &compare_insensitive($index_char, $top_char)) {
			$index_char = $top_char;
			$index_char =~ tr/a-z/A-Z/;
			$title = "= $index_char =";
		} else {
			undef $title;
		}

		if (defined($title)) {
			if ($title ne '= 0 =') {
				$fpwmenu->add_newline()
					|| die "Failed to add newline ("
							. $fpwmenu->error_message() . ").";
				&write_menu_footer($index_start, @lines);
				$index_start++;
			}
			&add_menu_page($title);
			$fpwmenu->add_newline()
				|| die "Failed to add newline" . &fpwerr($fpwmenu);
			$fpwmenu->add_indent_level(2)
				|| die "Failed to set indent level" . &fpwerr($fpwmenu);
		}

		&write_cross_reference($fpwmenu, $word_list[$i], $word_list[$i], 0);
		$fpwmenu->add_newline()
			|| die "Failed to add newline" . &fpwerr($fpwmenu) . ".";
	}

	$fpwmenu->add_newline()
		|| die "Failed to add newline" . &fpwerr($fpwmenu) . ".";
	&write_menu_footer($index_start, @lines);
}

sub write_menu_content($)
{
	my(@lines) = split(/[\t]/, $_[0]);
	my(@bullet) = ('*', '+', '-');
	my($index, $key, $title);

	$lines[0] =~ /^([0-9]):(.+)$/;
	$index = $1;
	$key = $2;

	if (0 < $index) {
		$fpwmenu->add_text(" " x $index);
		$title = $bullet[$index - 1] . " " if ($key !~ /^=/);
	}
	$title .= $key;

	&write_cross_reference($fpwmenu, $key, $title, 0);
	$fpwmenu->add_newline()
		|| die "Failed to add newline" . &fpwerr($fpwmenu) . ".";
}

sub add_menu_page($)
{
	my($title) = @_;
	my($key);

	print "writing menu entry <$title>...\n";

	# switch to the next context.
	$fpwmenu->new_context()
		|| die "Unable to add a new menu context" . &fpwerr($fpwmenu) . ".";

	# set the cross-reference tag.
	if ($title =~ /(TV Typewriters)/) {
		$key = $1;
	} else {
		$key = $title;
	}
	if (!defined($menues{$key})) {
		die "Unexpected menu topic \'$title\'.";
	}
	$fpwmenu->add_tag(sprintf("MENU%04d", $menues{$key}))
		|| die "Cannot add cross-reference tag" . &fpwerr($fpwmenu) . ".";

	# write the topic title.
	if (!$fpwmenu->add_keyword_start()
			|| !$fpwmenu->add_text($title)
			|| !$fpwmenu->add_keyword_end()
			|| !$fpwmenu->add_newline()) {
		die "Cannot write menu title \'$title\'", &fpwerr($fpwmenu) . ".";
	}
}

sub write_menu_footer($@)
{
	my($index, @lines) = @_;
	my($i);

	$fpwmenu->add_indent_level(2);
	for ($i = $index - 1; 0 <= $i; $i--) {
		$lines[$i] =~ /^[0-9]:([^\t]+)\t/;
		if (defined($menues{$1})) {
			$fpwmenu->add_text(' [');
			&write_cross_reference($fpwmenu, $1, 'Previous', 0);
			$fpwmenu->add_text(']');
			last;
		}
	}
	for ($i = $index + 1; $i < @lines; $i++) {
		$lines[$i] =~ /^[0-9]:([^\t]+)\t/;
		if (defined($menues{$1})) {
			$fpwmenu->add_text(' [');
			&write_cross_reference($fpwmenu, $1, 'Next', 0);
			$fpwmenu->add_text(']');
			last;
		}
	}
	$fpwmenu->add_newline();
	$fpwmenu->add_newline();
}
